from setuptools import Extension, find_packages
from setuptools.dist import Distribution
from pathlib import Path
import platform

try:
    from Cython.Build import cythonize
except ImportError:
    use_cython = False
    ext = 'c'
else:
    use_cython = True
    ext = 'pyx'

# Get all dxfeed c api c files to be compiled into separate lib
root_path = Path(__file__).resolve().parent
source_files_directory = root_path / 'dxfeed' / 'dxfeed-c-api' / 'src'
source_files_paths = [str(path) for path in source_files_directory.glob('*.c')]
libs = list()
include_dirs = [str(source_files_directory.parent / 'include'),
                str(source_files_directory)]
# Build dxfeed c library
dxfeed_c_lib_args = dict()
if platform.system() == 'Windows':
    source_files_paths.remove(str(source_files_directory / 'Linux.c'))
    libs.append('ws2_32')
else:
    source_files_paths.remove(str(source_files_directory / 'Win32.c'))
dxfeed_c_lib_args.update({'sources': source_files_paths,
                          'include_dirs': include_dirs})

if platform.system() == 'Darwin':
    dxfeed_c_lib_args.update({'macros': [('MACOSX', 'TRUE')]})

dxfeed_c = ('dxfeed_c', dxfeed_c_lib_args)

extensions = [Extension('dxfeed.core.utils.helpers', ['dxfeed/core/utils/helpers.' + ext],
                        include_dirs=include_dirs),
              Extension('dxfeed.core.utils.handler', ['dxfeed/core/utils/handler.' + ext],
                        include_dirs=include_dirs),
              Extension('dxfeed.core.listeners.listener', ['dxfeed/core/listeners/listener.' + ext],
                        include_dirs=include_dirs),
              Extension('dxfeed.core.DXFeedPy', ['dxfeed/core/DXFeedPy.' + ext], libraries=libs,
                        include_dirs=include_dirs)]

if use_cython:
    extensions = cythonize(extensions, language_level=3)

def build(setup_kwargs):
    setup_kwargs.update({
        'ext_modules': extensions,
        'zip_safe': False,
        'libraries': [dxfeed_c],
        'packages': find_packages(),
        'include_dirs': include_dirs
    })

def build_extensions():
    """
    Function for building extensions inplace for docs and tests
    :return:
    """
    build_params = {}
    build(build_params)
    dist = Distribution(attrs=build_params)
    build_clib_cmd = dist.get_command_obj('build_clib')
    build_clib_cmd.ensure_finalized()
    build_clib_cmd.run()
    build_ext_cmd = dist.get_command_obj('build_ext')
    build_ext_cmd.ensure_finalized()
    build_ext_cmd.inplace = 1
    build_ext_cmd.run()



from dxfeed.core.utils.helpers cimport *
from dxfeed.core.utils.handler cimport EventHandler
from collections import namedtuple

cdef class FuncWrapper:
    def __cinit__(self):
        self.func = NULL

    @staticmethod
    cdef FuncWrapper make_from_ptr(dxf_event_listener_t f):
        cdef FuncWrapper out = FuncWrapper()
        out.func = f
        return out


TRADE_COLUMNS = ['Symbol', 'Sequence', 'Price', 'ExchangeCode', 'Size', 'Tick', 'Change', 'DayVolume',
                 'DayTurnover', 'Direction', 'Time', 'TimeNanos', 'RawFlags', 'IsETH', 'Scope']
TradeTuple = namedtuple('Trade', ['symbol', 'sequence', 'price', 'exchange_code', 'size', 'tick', 'change',
                                  'day_volume', 'day_turnover', 'direction', 'time', 'time_nanos', 'raw_flags',
                                  'is_eth', 'scope'])
cdef void trade_default_listener(int event_type,
                                 dxf_const_string_t symbol_name,
                                 const dxf_event_data_t*data,
                                 int data_count, void*user_data) nogil:
    cdef dxf_trade_t* trades = <dxf_trade_t*> data

    with gil:
        py_data = <EventHandler> user_data
        events = [None] * data_count
        for i in range(data_count):
            events[i] = TradeTuple(symbol=unicode_from_dxf_const_string_t(symbol_name),
                                   sequence=trades[i].sequence,
                                   price=trades[i].price,
                                   exchange_code=unicode_from_dxf_const_string_t(&trades[i].exchange_code),
                                   size=trades[i].size,
                                   tick=trades[i].tick,
                                   change=trades[i].change,
                                   day_volume=trades[i].day_volume,
                                   day_turnover=trades[i].day_turnover,
                                   direction=trades[i].direction,
                                   time=trades[i].time,
                                   time_nanos=trades[i].time_nanos,
                                   raw_flags=trades[i].raw_flags,
                                   is_eth=trades[i].is_eth,
                                   scope=trades[i].scope)

        py_data.cython_internal_update_method(events)

QUOTE_COLUMNS = ['Symbol', 'Sequence', 'Time', 'TimeNanos', 'BidTime', 'BidExchangeCode', 'BidPrice', 'BidSize',
                 'AskTime', 'AskExchangeCode', 'AskPrice', 'AskSize', 'Scope']
QuoteTuple = namedtuple('Quote', ['symbol', 'sequence', 'time', 'time_nanos', 'bid_time', 'bid_exchange_code',
                                  'bid_price', 'bid_size', 'ask_time', 'ask_exchange_code', 'ask_price', 'ask_size',
                                  'scope'])
cdef void quote_default_listener(int event_type,
                                 dxf_const_string_t symbol_name,
                                 const dxf_event_data_t*data,
                                 int data_count,
                                 void*user_data) nogil:
    cdef dxf_quote_t*quotes = <dxf_quote_t*> data
    with gil:
        py_data = <EventHandler> user_data
        events = [None] * data_count
        for i in range(data_count):
            events[i] = QuoteTuple(symbol=unicode_from_dxf_const_string_t(symbol_name),
                                   sequence=quotes[i].sequence,
                                   time=quotes[i].time,
                                   time_nanos=quotes[i].time_nanos,
                                   bid_time=quotes[i].bid_time,
                                   bid_exchange_code=unicode_from_dxf_const_string_t(&quotes[i].bid_exchange_code),
                                   bid_price=quotes[i].bid_price,
                                   bid_size=quotes[i].bid_size,
                                   ask_time=quotes[i].ask_time,
                                   ask_exchange_code=unicode_from_dxf_const_string_t(&quotes[i].ask_exchange_code),
                                   ask_price=quotes[i].ask_price,
                                   ask_size=quotes[i].ask_size,
                                   scope=<int> quotes[i].scope)
        py_data.cython_internal_update_method(events)

SUMMARY_COLUMNS = ['Symbol', 'DayId', 'DayOpenPrice', 'DayHighPrice', 'DayLowPrice', 'DayClosePrice', 'PrevDayId',
                   'PrevDayClosePrice', 'PrevDayVolume', 'OpenInterest', 'RawFlags', 'ExchangeCode',
                   'DayClosePriceType', 'PrevDayClosePriceType', 'Scope']
SummaryTuple = namedtuple('Summary', ['symbol', 'day_id', 'day_open_price', 'day_high_price', 'day_low_price',
                                      'day_close_price', 'prev_day_id', 'prev_day_close_price', 'prev_day_volume',
                                      'open_interest', 'raw_flags', 'exchange_code', 'day_close_price_type',
                                      'prev_day_close_price_type', 'scope'])
cdef void summary_default_listener(int event_type, dxf_const_string_t symbol_name,
                                   const dxf_event_data_t*data, int data_count, void*user_data) nogil:
    cdef dxf_summary_t*summary = <dxf_summary_t*> data

    with gil:
        py_data = <EventHandler> user_data
        events = [None] * data_count
        for i in range(data_count):
            events[i] = SummaryTuple(symbol=unicode_from_dxf_const_string_t(symbol_name),
                                     day_id=summary[i].day_id,
                                     day_open_price=summary[i].day_open_price,
                                     day_high_price=summary[i].day_high_price,
                                     day_low_price=summary[i].day_low_price,
                                     day_close_price=summary[i].day_close_price,
                                     prev_day_id=summary[i].prev_day_id,
                                     prev_day_close_price=summary[i].prev_day_close_price,
                                     prev_day_volume=summary[i].prev_day_volume,
                                     open_interest=summary[i].open_interest,
                                     raw_flags=summary[i].raw_flags,
                                     exchange_code=unicode_from_dxf_const_string_t(&summary[i].exchange_code),
                                     day_close_price_type=summary[i].day_close_price_type,
                                     prev_day_close_price_type=summary[i].prev_day_close_price_type,
                                     scope=summary[i].scope)
        py_data.cython_internal_update_method(events)

PROFILE_COLUMNS = ['Symbol', 'Beta', 'EPS', 'DivFreq', 'ExdDivAmount', 'ExdDivDate', '52HighPrice', '52LowPrice',
                   'Shares', 'FreeFloat', 'HighLimitPrice', 'LowLimitPrice', 'HaltStartTime', 'HaltEndTime',
                   'Description', 'RawFlags', 'StatusReason', 'TradingStatus', 'ShortSaleRestriction']
ProfileTuple = namedtuple('Profile', ['symbol', 'beta', 'eps', 'div_freq', 'exd_div_amount', 'exd_div_date',
                                      'high_price', 'low_price', 'shares', 'free_float', 'high_limit_price',
                                      'low_limit_price', 'halt_start_time', 'halt_end_time', 'description', 'raw_flags',
                                      'status_reason', 'trading_status', 'ssr'])
cdef void profile_default_listener(int event_type,
                                   dxf_const_string_t symbol_name,
                                   const dxf_event_data_t*data,
                                   int data_count,
                                   void*user_data)  nogil:
    cdef dxf_profile_t*p = <dxf_profile_t*> data
    with gil:
        py_data = <EventHandler> user_data
        events = [None] * data_count
        for i in range(data_count):
            events[i] = ProfileTuple(symbol=unicode_from_dxf_const_string_t(symbol_name),
                                     beta=p[i].beta,
                                     eps=p[i].eps,
                                     div_freq=p[i].div_freq,
                                     exd_div_amount=p[i].exd_div_amount,
                                     exd_div_date=p[i].exd_div_date,
                                     high_price=p[i]._52_high_price,
                                     low_price=p[i]._52_low_price,
                                     shares=p[i].shares,
                                     free_float=p[i].free_float,
                                     high_limit_price=p[i].high_limit_price,
                                     low_limit_price=p[i].low_limit_price,
                                     halt_start_time=p[i].halt_start_time,
                                     halt_end_time=p[i].halt_end_time,
                                     description=unicode_from_dxf_const_string_t(p[i].description),
                                     raw_flags=p[i].raw_flags,
                                     status_reason=unicode_from_dxf_const_string_t(p[i].status_reason),
                                     trading_status=p[i].trading_status,
                                     ssr=p[i].ssr)
        py_data.cython_internal_update_method(events)

TIME_AND_SALE_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'ExchangeCode', 'Price', 'Size', 'BidPrice',
                         'AskPrice', 'ExchangeSaleConditions', 'RawFlags', 'Buyer', 'Seller', 'Side', 'Type',
                         'IsValidTick', 'IsEthTrade', 'TradeThroughExempt', 'IsSpreadLeg', 'Scope']
TnSTuple = namedtuple('TnS', ['symbol', 'event_flags', 'index', 'time', 'exchange_code', 'price', 'size', 'bid_price',
                              'ask_price', 'exchange_sale_conditions', 'raw_flags', 'buyer', 'seller', 'side', 'type',
                              'is_valid_tick', 'is_eth_trade', 'trade_through_exempt', 'is_spread_leg', 'scope'])
cdef void time_and_sale_default_listener(int event_type,
                                         dxf_const_string_t symbol_name,
                                         const dxf_event_data_t*data,
                                         int data_count,
                                         void*user_data) nogil:
    cdef dxf_time_and_sale_t*tns = <dxf_time_and_sale_t*> data
    with gil:
        py_data = <EventHandler> user_data
        events = [None] * data_count
        for i in range(data_count):
            events[i] = TnSTuple(symbol=unicode_from_dxf_const_string_t(symbol_name),
                                 event_flags=tns[i].event_flags,
                                 index=tns[i].index,
                                 time=tns[i].time,
                                 exchange_code=unicode_from_dxf_const_string_t(&tns[i].exchange_code),
                                 price=tns[i].price,
                                 size=tns[i].size,
                                 bid_price=tns[i].bid_price,
                                 ask_price=tns[i].ask_price,
                                 exchange_sale_conditions= \
                                     unicode_from_dxf_const_string_t(tns[i].exchange_sale_conditions),
                                 raw_flags=tns[i].raw_flags,
                                 buyer=unicode_from_dxf_const_string_t(tns[i].buyer),
                                 seller=unicode_from_dxf_const_string_t(tns[i].seller),
                                 side=tns[i].side,
                                 type=tns[i].type,
                                 is_valid_tick=tns[i].is_valid_tick,
                                 is_eth_trade=tns[i].is_eth_trade,
                                 trade_through_exempt=tns[i].trade_through_exempt,
                                 is_spread_leg=tns[i].is_spread_leg,
                                 scope=tns[i].scope)
        py_data.cython_internal_update_method(events)

CANDLE_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'Sequence', 'Count', 'Open', 'High', 'Low', 'Close',
                  'Volume', 'VWap', 'BidVolume', 'AskVolume', 'OpenInterest', 'ImpVolatility']
CandleTuple = namedtuple('Candle', ['symbol', 'event_flags', 'index', 'time', 'sequence', 'count', 'open', 'high',
                                    'low', 'close', 'volume', 'vwap', 'bid_volume', 'ask_volume', 'open_interest',
                                    'imp_volatility'])
cdef void candle_default_listener(int event_type,
                                  dxf_const_string_t symbol_name,
                                  const dxf_event_data_t*data,
                                  int data_count,
                                  void*user_data) nogil:
    cdef dxf_candle_t*candle = <dxf_candle_t*> data
    with gil:
        py_data = <EventHandler> user_data
        events = [None] * data_count
        for i in range(data_count):
            events[i] = CandleTuple(symbol=unicode_from_dxf_const_string_t(symbol_name),
                                    event_flags=candle[i].event_flags,
                                    index=candle[i].index,
                                    time=candle[i].time,
                                    sequence=candle[i].sequence,
                                    count=candle[i].count,
                                    open=candle[i].open,
                                    high=candle[i].high,
                                    low=candle[i].low,
                                    close=candle[i].close,
                                    volume=candle[i].volume,
                                    vwap=candle[i].vwap,
                                    bid_volume=candle[i].bid_volume,
                                    ask_volume=candle[i].ask_volume,
                                    open_interest=candle[i].open_interest,
                                    imp_volatility=candle[i].imp_volatility)
        py_data.cython_internal_update_method(events)

ORDER_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'TimeNanos', 'Sequence', 'Price', 'Size', 'Count', 'Scope',
                 'Side', 'ExchangeCode', 'Source', 'MarketMaker', 'SpreadSymbol']
OrderTuple = namedtuple('Order', ['symbol', 'event_flags', 'index', 'time', 'time_nanos', 'sequence', 'price', 'size',
                                  'count', 'scope', 'side',  'exchange_code', 'source', 'market_maker', 'spread_symbol'])
cdef void order_default_listener(int event_type,
                                 dxf_const_string_t symbol_name,
                                 const dxf_event_data_t*data,
                                 int data_count,
                                 void*user_data) nogil:
    cdef dxf_order_t*order = <dxf_order_t*> data
    with gil:
        py_data = <EventHandler> user_data
        events = [None] * data_count
        for i in range(data_count):
            events[i] = OrderTuple(symbol=unicode_from_dxf_const_string_t(symbol_name),
                                   event_flags=order[i].event_flags,
                                   index=order[i].index,
                                   time=order[i].time,
                                   time_nanos=order[i].time_nanos,
                                   sequence=order[i].sequence,
                                   price=order[i].price,
                                   size=order[i].size,
                                   count=order[i].count,
                                   scope=order[i].scope,
                                   side=order[i].side,
                                   exchange_code=unicode_from_dxf_const_string_t(&order[i].exchange_code),
                                   source=unicode_from_dxf_const_string_t(&order[i].source[DXF_RECORD_SUFFIX_SIZE]),
                                   market_maker=unicode_from_dxf_const_string_t(order[i].market_maker),
                                   spread_symbol=unicode_from_dxf_const_string_t(order[i].spread_symbol))
        py_data.cython_internal_update_method(events)

GREEKS_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'Price', 'Volatility', 'Delta', 'Gamma', 'Theta', 'Rho',
                  'Vega']
GreekTuple = namedtuple('Greek', ['symbol', 'event_flags', 'index', 'time', 'price', 'volatility', 'delta', 'gamma',
                                  'theta', 'rho', 'vega'])
cdef void greeks_default_listener(int event_type,
                                  dxf_const_string_t symbol_name,
                                  const dxf_event_data_t*data,
                                  int data_count,
                                  void*user_data) nogil:
    cdef dxf_greeks_t*greeks = <dxf_greeks_t*> data
    with gil:
        py_data = <EventHandler> user_data
        events = [None] * data_count
        for i in range(data_count):
            events[i] = GreekTuple(symbol=unicode_from_dxf_const_string_t(symbol_name),
                                   event_flags=greeks[i].event_flags,
                                   index=greeks[i].index,
                                   time=greeks[i].time,
                                   price=greeks[i].price,
                                   volatility=greeks[i].volatility,
                                   delta=greeks[i].delta,
                                   gamma=greeks[i].gamma,
                                   theta=greeks[i].theta,
                                   rho=greeks[i].rho,
                                   vega=greeks[i].vega)
        py_data.cython_internal_update_method(events)

THEO_PRICE_COLUMNS = ['Symbol', 'Time', 'Price', 'UnderlyingPrice', 'Delta', 'Gamma', 'Dividend', 'Interest']
TheoPriceTuple = namedtuple('TheoPrice', ['symbol', 'time', 'price', 'underlying_price', 'delta', 'gamma', 'dividend',
                                          'interest'])
cdef void theo_price_default_listener(int event_type,
                                      dxf_const_string_t symbol_name,
                                      const dxf_event_data_t*data,
                                      int data_count,
                                      void*user_data) nogil:
    cdef dxf_theo_price_t*theo_price = <dxf_theo_price_t*> data
    with gil:
        py_data = <EventHandler> user_data
        events = [None] * data_count
        for i in range(data_count):
            events[i] = TheoPriceTuple(symbol=unicode_from_dxf_const_string_t(symbol_name),
                                       time=theo_price[i].time,
                                       price=theo_price[i].price,
                                       underlying_price=theo_price[i].underlying_price,
                                       delta=theo_price[i].delta,
                                       gamma=theo_price[i].gamma,
                                       dividend=theo_price[i].dividend,
                                       interest=theo_price[i].interest)
        py_data.cython_internal_update_method(events)

UNDERLYING_COLUMNS = ['Symbol', 'Volatility', 'FrontVolatility', 'BackVolatility', 'PutCallRatio']
UnderlyingTuple = namedtuple('Underlying', ['symbol', 'volatility', 'front_volatility', 'back_volatility',
                                            'put_call_ratio'])
cdef void underlying_default_listener(int event_type,
                                      dxf_const_string_t symbol_name,
                                      const dxf_event_data_t*data,
                                      int data_count,
                                      void*user_data) nogil:
    cdef dxf_underlying_t*underlying = <dxf_underlying_t*> data
    with gil:
        py_data = <EventHandler> user_data
        events = [None] * data_count
        for i in range(data_count):
            events[i] = UnderlyingTuple(symbol=unicode_from_dxf_const_string_t(symbol_name),
                                        volatility=underlying[i].volatility,
                                        front_volatility=underlying[i].front_volatility,
                                        back_volatility=underlying[i].back_volatility,
                                        put_call_ratio=underlying[i].put_call_ratio)
        py_data.cython_internal_update_method(events)

SERIES_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'Sequence', 'Expiration', 'Volatility',
                  'PutCallRatio', 'ForwardPrice', 'Dividend', 'Interest']
SeriesTuple = namedtuple('Series', ['symbol', 'event_flags', 'index', 'time', 'sequence', 'expiration', 'volatility',
                                    'put_call_ratio', 'forward_price', 'dividend', 'interest'])
cdef void series_default_listener(int event_type,
                                  dxf_const_string_t symbol_name,
                                  const dxf_event_data_t*data,
                                  int data_count,
                                  void*user_data) nogil:
    cdef dxf_series_t*series = <dxf_series_t*> data
    with gil:
        py_data = <EventHandler> user_data
        events = [None] * data_count
        for i in range(data_count):
            events[i] = SeriesTuple(symbol=unicode_from_dxf_const_string_t(symbol_name),
                                    event_flags=series[i].event_flags,
                                    index=series[i].index,
                                    time=series[i].time,
                                    sequence=series[i].sequence,
                                    expiration=series[i].expiration,
                                    volatility=series[i].volatility,
                                    put_call_ratio=series[i].put_call_ratio,
                                    forward_price=series[i].forward_price,
                                    dividend=series[i].dividend,
                                    interest=series[i].interest)
        py_data.cython_internal_update_method(events)

CONFIGURATION_COLUMNS = ['Symbol', 'Version', 'Object']
ConfigurationTuple = namedtuple('Configuration', ['symbol', 'version', 'object'])
cdef void configuration_default_listener(int event_type,
                                         dxf_const_string_t symbol_name,
                                         const dxf_event_data_t*data,
                                         int data_count,
                                         void*user_data) nogil:
    cdef dxf_configuration_t*config = <dxf_configuration_t*> data
    with gil:
        py_data = <EventHandler> user_data
        events = [None] * data_count
        for i in range(data_count):
            events[i] = ConfigurationTuple(symbol=unicode_from_dxf_const_string_t(symbol_name),
                                           version=config[i].version,
                                           object=unicode_from_dxf_const_string_t(config[i].object))
        py_data.cython_internal_update_method(events)



from typing import Iterable, Union, Optional
from datetime import datetime
from warnings import simplefilter

from dxfeed.core import DXFeedPy as dxp
from dxfeed.core.utils.handler import DefaultHandler
import dxfeed.wrappers.class_utils as cu


class Subscription(object):
    """
    Class for subscription management. Recommended to be created only via create_subscription method in Endpoint class.
    Also stores incoming events

    Attributes
    ----------
    event_type: str
        One of possible event types: 'Trade', 'Quote', 'Summary', 'Profile', 'Order', 'TimeAndSale', 'Candle',
        'TradeETH', 'SpreadOrder', 'Greeks', 'TheoPrice', 'Underlying', 'Series', 'Configuration' or ''
    symbols: Iterable
        Symbols of current subscription.

    Note
    ----
    Some event types (e.g. Candle) support only timed subscription.

    """
    def __init__(self, connection, event_type: str, date_time: Union[str, datetime], exact_format: bool = True):
        """

        Parameters
        ----------
        connection: dxfeed.core.DXFeedPy.ConnectionClass
            Core class written in cython, that handle connection related details on the low level
        event_type: str
            One of possible event types: 'Trade', 'Quote', 'Summary', 'Profile', 'Order', 'TimeAndSale', 'Candle',
            'TradeETH', 'SpreadOrder', 'Greeks', 'TheoPrice', 'Underlying', 'Series', 'Configuration' or ''
        date_time: str or datetime.datetime
            If present timed subscription will be created (conflated stream). For sting date format is following:
            %Y-%m-%d %H:%M:%S.%f. If None - stream subscription will be created (non-conflated). Default - None.
        exact_format: bool
            If False no warning will be thrown in case of incomplete date_time parameter. Default - True
        """
        self.__event_type = event_type
        if date_time is None:
            self.__sub = dxp.dxf_create_subscription(cc=connection,
                                                     event_type=event_type)
        else:
            date_time = cu.handle_datetime(date_time, fmt='%Y-%m-%d %H:%M:%S.%f', exact_format=exact_format)
            timestamp = int(date_time.timestamp() * 1000)
            self.__sub = dxp.dxf_create_subscription_timed(cc=connection,
                                                           event_type=event_type,
                                                           time=timestamp)

    def __del__(self):
        self.close_subscription()

    @property
    def event_type(self):
        return self.__event_type

    @property
    def symbols(self):
        return dxp.dxf_get_symbols(self.__sub)

    def add_symbols(self, symbols: Union[str, Iterable[str]]):
        """
        Method to add symbol. Supports addition of one symbol as a string as well as list of symbols.
        If no event handler was set, DefaultHandler will be initialized.

        Parameters
        ----------
        symbols: str, Iterable
            Symbols to add. Previously added and remained symbols are ignored on C-API level

        Returns
        -------
        self: Subscription
        """
        self._attach_default_listener()
        dxp.dxf_add_symbols(sc=self.__sub, symbols=cu.to_iterable(symbols))
        return self

    def remove_symbols(self, symbols: Optional[Union[str, Iterable[str]]] = None):
        """
        Method removes symbols from subscription. If no symbols provided removes all symbols

        Parameters
        ----------
        symbols: str, Iterable
            One ticker or list of tickers to remove from subscription

        Returns
        -------
        self: Subscription
        """
        if symbols:
            dxp.dxf_remove_symbols(self.__sub, symbols=cu.to_iterable(symbols))
        else:
            dxp.dxf_clear_symbols(self.__sub)
        return self

    def close_subscription(self):
        """
        Method to close subscription. All received data will remain in the object.
        """
        dxp.dxf_close_subscription(sc=self.__sub)

    def _attach_default_listener(self):
        """
        Method to attach default listener. If event handler was not previously set, DefaultHandler will be initialized.

        Returns
        -------
        self: Subscription
        """
        if not self.get_event_handler():
            self.set_event_handler(DefaultHandler())
            simplefilter(action='ignore', category=FutureWarning)

        dxp.dxf_attach_listener(self.__sub)

        simplefilter(action='default', category=FutureWarning)

        return self

    def _detach_listener(self):
        """
        Removes listener so new events won't be received

        Returns
        -------
        self: Subscription
        """
        dxp.dxf_detach_listener(self.__sub)
        return self

    def get_event_handler(self):
        """
        Method to get event handler. If no handlers passed previously returns None.

        Returns
        -------
        handler: EventHandler
        """
        return self.__sub.get_event_handler()

    def set_event_handler(self, handler: dxp.EventHandler):
        """
        Method to set the handler.

        Parameters
        ----------
        handler: EventHandler
            Event handler with `update` method defined.

        Returns
        -------
        self: Subscription
        """
        self.__sub.set_event_handler(handler)
        return self



from typing import Union, Iterable
import collections
from warnings import warn
from datetime import datetime
import pandas as pd


def to_iterable(value: Union[str, Iterable[str]]) -> Iterable[str]:
    """
    Function to correctly wrap string into iterable object

    Parameters
    ----------
    value: str, iterable
        String or iterable object

    Returns
    -------
    : iterable
    """
    if isinstance(value, str):
        value = [value]
    if not isinstance(value, collections.abc.Iterable):
        raise TypeError(f'Expected string or iterable type, got {type(value)}')
    return value


def handle_datetime(date_time: Union[str, datetime], fmt: str = '%Y-%m-%d %H:%M:%S.%f', exact_format: bool = True):
    """
    Function to convert string of date and time to datetime object.

    Parameters
    ----------
    date_time: str, datetime.datetime
        Datetime to convert
    fmt: str
        Format of expected datetime
    exact_format: bool
        If False no warning will be thrown in case of incomplete date_time parameter. Default - True

    Returns
    -------
    date_time: datetime.datetime
        date_time argument converted to datetime.datetime object.
    """
    if isinstance(date_time, str):
        try:
            date_time = datetime.strptime(date_time, fmt)
        except ValueError:
            try:
                date_time = pd.to_datetime(date_time, format=fmt, infer_datetime_format=True)
                if exact_format:
                    warn(Warning(f'Datetime argument does not exactly match {fmt} format,' +
                                 ' date was parsed automatically as ' +
                                 date_time.strftime(format=fmt)))
            except ValueError:
                raise ValueError(f'Datetime should use {fmt} format!')
    if not isinstance(date_time, datetime):
        raise TypeError(f'Datetime has incorrect type of {type(date_time)}, should be string or datetime')
    return date_time



from warnings import warn
from dxfeed.core.utils.data_class import DequeWithLock as deque_wl
import pandas as pd
from typing import Iterable


cdef class EventHandler:
    """
    Master class for user defined event handlers. `update` method should be considered as abstract.

    Attributes
    ----------
    columns: list
        After attaching listener the field contains one-word descriptors of the values in upcoming event the order
        coincides
    """
    def __init__(self):
        self.columns = list()

    cdef void cython_internal_update_method(self, event) nogil:
        with gil:
            self.update(event)

    def update(self, event: Iterable):
        """
        Method that is called, when event arrives to related Subscription. Currently (Cython version 0.29.17),
        abstract methods are not implemented, so this implementation is sort of stub method.

        Parameters
        ----------
        event: list
            List of various data specific to certain event type
        """
        warn(Warning('You have not implemented update method in your EventHandler, that is called, when event comes!'))

cdef class DefaultHandler(EventHandler):
    def __init__(self, data_len: int=100000):
        """
        The class implements one possible event handler, which is considered as default. This class just stores upcoming
        events in custom deque with thread lock and has methods to get data as list or as pandas.DataFrame.

        Attributes
        ----------
        data_len: int
            The length of internal DequeWithLock object. Default is 100000.
        columns: list
            After attaching listener the field contains one-word descriptors of the values in upcoming event the order
            coincides.
        """
        super().__init__()
        self.__data = deque_wl(maxlen=data_len)

    def update(self, event: Iterable):
        """
        Utility method that is called by underlying Cython level when new event is received. Stores events in
        DequeWithLock.

        Parameters
        ----------
        event: list
            List of various data specific to certain event type.
        """
        self.__data.safe_append(*event)

    def get_list(self, keep: bool=True):
        """
        Method to get data stored in DequeWithLock as list.

        Parameters
        ----------
        keep: bool
            When False clears internal DequeWithLock object after call. Default True.

        Returns
        -------
        data: list
            List of received events.
        """
        data = self.__data.safe_get(keep=keep)
        return data

    def get_dataframe(self, keep: bool=True):
        """
        Returns received data as pandas.DataFrame object.

        Warning
        -------
        There are no restriction on attaching one event handler to several subscriptions. In this case, method will
        work correctly if different subscriptions have same event type. Otherwise error will occur.

        Parameters
        ----------
        keep: bool
            When False clears internal DequeWithLock object after call. Default True.

        Returns
        -------
        df: pandas.DataFrame
            Dataframe with received events.
        """
        df_data = self.get_list(keep=keep)


        df = pd.DataFrame(df_data, columns=self.columns)
        time_columns = (column for column in df.columns
                        if 'Time' in column
                        and 'TimeNanos' not in column)
        for column in time_columns:
            df.loc[:, column] = df.loc[:, column].values.astype('datetime64[ms]')
        if 'TimeNanos' in df.columns and 'Time' in df.columns:
            df.loc[:, 'Time'] += df.loc[:, 'TimeNanos'].values.astype('timedelta64[ns]')
            df.drop(columns=['TimeNanos'], inplace=True)
        return df



from dxfeed.wrappers.endpoint import Endpoint
from dxfeed.core.utils.handler import EventHandler, DefaultHandler
import pkg_resources
import toml
from pathlib import Path


try:
    __version__ = pkg_resources.get_distribution('dxfeed').version
except pkg_resources.DistributionNotFound:
    pyproject = toml.load(Path(__file__).parents[1] / 'pyproject.toml')
    __version__ = pyproject['tool']['poetry']['version']



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dxfeed',
 'dxfeed.core',
 'dxfeed.core.listeners',
 'dxfeed.core.pxd_include',
 'dxfeed.core.utils',
 'dxfeed.wrappers']

package_data = \
{'': ['*'],
 'dxfeed': ['dxfeed-c-api/*',
            'dxfeed-c-api/include/*',
            'dxfeed-c-api/include/pthreads/*',
            'dxfeed-c-api/src/*']}

install_requires = \
['pandas>=0.25.1,<0.26.0', 'toml>=0.10.0,<0.11.0']

extras_require = \
{'docs': ['jupyter>=1.0.0,<2.0.0']}

setup_kwargs = {
    'name': 'dxfeed',
    'version': '0.5.0',
    'description': 'DXFeed Python API via C API',
    'long_description': "# dxfeed package\n\n[![PyPI](https://img.shields.io/pypi/v/dxfeed)](https://pypi.org/project/dxfeed/)\n[![Documentation Status](https://readthedocs.org/projects/dxfeed/badge/?version=latest)](https://dxfeed.readthedocs.io/en/latest/?badge=latest)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/dxfeed)](https://pypi.org/project/dxfeed/)\n[![PyPI - Wheel](https://img.shields.io/pypi/wheel/dxfeed)](https://pypi.org/project/dxfeed/)\n[![PyPI - License](https://img.shields.io/pypi/l/dxfeed)](https://github.com/dxFeed/dxfeed-python-api/blob/master/LICENSE)\n[![Test workflow](https://github.com/dxFeed/dxfeed-python-api/workflows/Test%20package/badge.svg)](https://github.com/dxFeed/dxfeed-python-api/actions)\n\n\n\nThis package provides access to [dxFeed](https://www.dxfeed.com/) streaming data.\nThe library is build as a thin wrapper over [dxFeed C-API library](https://github.com/dxFeed/dxfeed-c-api).\nWe use [Cython](https://cython.org/) in this project as it combines flexibility, reliability and\nusability in writing C extensions.\n\nThe design of the dxfeed package allows users to write any logic related to events in python as well as \nextending lower level Cython functionality. Moreover, one may start working with the API using the default \nvalues like function arguments or a default event handler.\n\nDocumentation: [dxfeed.readthedocs.io](https://dxfeed.readthedocs.io/en/latest/)\n\nPackage distribution: [pypi.org/project/dxfeed](https://pypi.org/project/dxfeed/)\n\n## Installation\n\n**Requirements:** python >= 3.6\n\nInstall package via PyPI:\n\n```python\npip3 install dxfeed\n``` \n\n## Installation from sources\n\nReminder: initialize and pull git submodule after cloning the repo:\n```bash\ngit submodule init\ngit submodule update\n``` \n\nTo install dxfeed from source you need Poetry. It provides a custom installer.\nThis is the recommended way of installing poetry according to [documentation](https://python-poetry.org/docs/)\n\nFor macOS / Linux / Windows (with bash):\n\n```bash\ncurl -sSL https://raw.githubusercontent.com/python-poetry/poetry/master/get-poetry.py | python\n```\n\nIn the project root directory (same one where you found this file after\ncloning the git repo), execute:\n\n```bash\npoetry install \n```\n\nBy default package is installed in \n[development mode](https://pip.pypa.io/en/latest/reference/pip_install.html#editable-installs). To rebuild \nC extensions, after editing .pyx files:\n\n```bash\npoetry run task build_inplace  # build c extensions\n```\n\n## Basic usage\n\nFollowing steps should be performed:\n\n* Import\n* Create Endpoint\n* Create Subscription\n* Attach event handler\n* Add tickers\n* Finally close subscription and connection \n\n### Import package\n\n```python\nimport dxfeed as dx\nfrom datetime import datetime  # for timed subscription\n```\n\n### Configure and create connection with Endpoint class\nCreate instance of Endpoint class which will connect provided address. \n\n\n```python\nendpoint = dx.Endpoint('demo.dxfeed.com:7300')\n```\n\nEndpoint instance contains information about the connection, e.g. connection address or status\n\n\n```python\nprint(f'Connected address: {endpoint.address}')\nprint(f'Connection status: {endpoint.connection_status}')\n```\n\n```text\nConnected address: demo.dxfeed.com:7300\nConnection status: Connected and authorized\n```\n\n### Configure and create subscription\nYou should specify event type. For timed subscription (conflated stream) you should also provide time to start subscription from.\n\n\n```python\ntrade_sub = endpoint.create_subscription('Trade')\n```\n\n**Attach default or custom event handler** - class that process incoming events. For details about custom\nevent handler look into `CustomHandlerExample.ipynb` jupyter notebook in `exapmles` folder of this repository.\n\n\n```python\ntrade_handler = dx.DefaultHandler()\ntrade_sub = trade_sub.set_event_handler(trade_handler)\n```\n\n**Add tikers** you want to receive events for\n\n\n```python\ntrade_sub = trade_sub.add_symbols(['C', 'AAPL'])\n```\n\nFor timed subscription you may provide either datetime object or string. String might be incomplete, in \nthis case you will get warning with how your provided date parsed automatically. \n\n\n```python\ntns_sub = endpoint.create_subscription('TimeAndSale', date_time=datetime.now()) \\\n                  .add_symbols(['AMZN'])\n```\n\n\n```python\ncandle_sub = endpoint.create_subscription('Candle', date_time='2020-04-16 13:05')\ncandle_sub = candle_sub.add_symbols(['AAPL', 'MSFT'])\n```\n\nWe didn't provide subscriptions with event handlers. In such a case DefaultHandler is initiated automatically.\nOne may get it with `get_event_handler` method.\n\n```python\ntns_handler = tns_sub.get_event_handler()\ncandle_handler = candle_sub.get_event_handler()\n```\n\n#### Subscription instance properties\n\n\n```python\nprint(f'Subscription event type: {tns_sub.event_type}')\nprint(f'Subscription symbols: {candle_sub.symbols}')\n```\n\n```text\nSubscription event type: TimeAndSale\nSubscription symbols: ['AAPL', 'MSFT']\n```\n\n### Access data\nIn DefaultHandler the data is stored as deque. Its length may be configured, by default 100000 events.\n\n```python\ncandle_handler.get_list()\n```\n\n### Close connection\n\n\n```python\nendpoint.close_connection()\nprint(f'Connection status: {endpoint.connection_status}')\n```\n\n```text\nConnection status: Not connected\n```\n\n### Transform data to pandas DataFrame\n\nDefaultHandler has `get_dataframe` method, which allows you to get pandas.DataFrame object with events as rows.\n\n```python\ntrade_df = trade_handler.get_dataframe()\ntns_df = tns_handler.get_dataframe()\ncandle_df = candle_handler.get_dataframe()\n```\n",
    'author': 'Index Management Team',
    'author_email': 'im@dxfeed.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



from dxfeed.core import DXFeedPy as dxp
from dxfeed.wrappers.subscription import Subscription
from datetime import datetime
from typing import Union


class Endpoint(object):
    """
    Class for connection management. After successful creation Instance will be connected to server
    with provided credentials

    Attributes
    ----------
    connection_status: str
        Status of current connection
    address: str
        Current connection endpoint address

    """
    def __init__(self, connection_address: str = 'demo.dxfeed.com:7300', connect: bool = True):
        """
        Parameters
        ----------
        connection_address: str
            One of possible connection addresses:

            - the single address: `host:port` or just `host`
            - address with credentials: `host:port[username=xxx,password=yyy]`
            - multiple addresses: `(host1:port1)(host2)(host3:port3[username=xxx,password=yyy])`

            Default: demo.dxfeed.com:7300
        connect: bool
            When True `connect` method  is called during instance creation. Default - True
        """
        self.__con_address = connection_address
        self.__connection = dxp.ConnectionClass()
        if connect:
            self.connect()

    def __del__(self):
        self.close_connection()

    @property
    def connection_status(self):
        return dxp.dxf_get_current_connection_status(self.__connection, return_str=True)

    @property
    def address(self):
        return self.__con_address

    def connect(self, reconnect: bool = True):
        """
        Creates connection. If connection status differs from "Not connected" and `reconnect` is False, does nothing

        Parameters
        ----------
        reconnect: bool
            When True closes previous connection. Default - True

        Returns
        -------
        self: Endpoint
        """
        if reconnect:
            dxp.dxf_close_connection(self.__connection)

        con_status = dxp.dxf_get_current_connection_status(self.__connection, return_str=True)

        if con_status == 'Not connected':
            self.__connection = dxp.dxf_create_connection(self.address)

        return self

    def create_subscription(self, event_type: str, date_time: Union[str, datetime] = None):
        """
        Method creates certain event type subscription and returns Subscription class

        Parameters
        ----------
        event_type: str
            One of possible event types: 'Trade', 'Quote', 'Summary', 'Profile', 'Order', 'TimeAndSale', 'Candle',
            'TradeETH', 'SpreadOrder', 'Greeks', 'TheoPrice', 'Underlying', 'Series', 'Configuration' or ''
        date_time: str or datetime.datetime
            If present timed subscription will be created (conflated stream). For sting date format is following:
            %Y-%m-%d %H:%M:%S.%f. If None - stream subscription will be created. Default - None.

        Note
        ----
        Some event types (e.g. Candle) support only timed subscription.

        Returns
        -------
        subscription: Subscription
            Subscription class related to current connection
        """
        con_status = dxp.dxf_get_current_connection_status(self.__connection, return_str=False)
        if con_status == 0 or con_status == 2:
            raise ValueError('Connection is not established')
        subscription = Subscription(connection=self.__connection,
                                    event_type=event_type,
                                    date_time=date_time)
        return subscription

    def close_connection(self):
        """
        Method to close connections and all related subscriptions.
        """
        dxp.dxf_close_connection(self.__connection)
